/*
   Servo Controller. Atmel ATmega48/88/168 based.
   ELM Chan + Alexander A. Karsakbayev + Arisov A.V.
 


	Version v3.6
	Clock: 16 MHz 
	AtMega168


-------------------------------------------
    Version_v3.1
-  EEPROM
- ""  
-    ,
- E-Stop Input
-  Error (2  )
-     )
-------------------------------------------- 

    Version_v3.2
-  WatcDog   125 (     "" )

-     B_PWM_OUT1  B_PWM_OUT2 (..    
        ElmChan http://elm-chan.org/works/smc/smc.png)
    
-    /  Direct_L  Direct_R (  
   Dir_L  Dir_R  )  

 -        

-          
       , 
      ,    
 (    Error)

 -------------------------------------------
     Version_v3.3
-        

 -------------------------------------------
     Version_v3.4 (for AtMega48 = 16MHz)
-    AtMega48 (   RAM,   "")
-      +/-240 . (     ) 

  -------------------------------------------
     Version_v3.5 (for AtMega48 = 20MHz)
-       20 (UART & PWM)
-     19,5

  -------------------------------------------
     Version_v3.6 (for AtMega168 = 16MHz)
-   AtMega168

   
 */



.include "m168def.inc"


;-------------------------------------
; PORT B
.equ	Mode_Lim	= 3		;  Mode_Lim (   )
	;   () (Inp=1) - -  ()
	;    (Inp=0) - -  ()  
.equ	Lim_L		= 4		;  Limit_L
.equ	Lim_R		= 5		;  Limit_R

; PORT C
.equ	Dir_L	 	= 5		;  L (    )
.equ	Dir_R 		= 4		;  R
.equ	DIRECT		= 3		;  , 
.equ	HL_Error 	= 2		;  Error

; PORT D 
.equ	Enc_A	 	= 4		;   
.equ	Enc_B	 	= 5		;  B 



;-------------------------------------
.def	_Flags		= r23	;
.def	Flags2		= r25	;
.def	W			= r22	;   / EEPROM_RAM_Majorization	

.def	uart_r		= r24 	;  ,    UART_OUT

.def	_0			= r15 	; 

.def	_PvEnc		= r14 	;    
.def	_PvDir		= r13 	;    

.def	_PosH		= r12 	; 24-bit    
.def	_PosM		= r11 	;
.def	_PosL		= r10 	;

.def	_CtDiv83	= r7  	;  1/83 (1kHz)



;-------------------------------------
; _Flags
.equ	full_bit		=0	;
.equ	command_bit		=1 	;      

.equ	wait_cmd_bit	=2	;  
.equ	wait_size_bit	=3	;  


; _Flags2
.equ	EE_WR1			=0	;   
.equ	EE_WR2			=1	;  EEPROM  
.equ	EE_WR3			=2	; / EEPROM_RAM_Majorization

.equ	Lim1_L			=0	;   
.equ	Lim2_L			=1	;    
.equ	Lim1_R			=2	;  
.equ	Lim2_R			=3	;     
.equ	Mode1_Lim		=4	;   (  )
.equ	Mode2_Lim		=5	; ( )
.equ	StepCNC			=6	;     
.equ	Error			=7	;  ""
;-------------------------------------

.dseg
		.org	0x100

VelError: 		.byte  	2	;  	
VelError200: 	.byte  	2	;   (200Hz)
Velocity: 		.byte  	2 	; 	

CtPos:			.byte	3	; 24-bit  		g/j	mode 3
CtSub:			.byte	2	; Sub command   		s	mode 0/1/2
PvInt:			.byte	2	; Integration register
PvPos:			.byte	2	; 16-bit Velocity detection register
OvTmr:			.byte	2	; Torque limit timer

Parms:
LimSpd:			.byte	2	;P0,Velocity limit		Integer
GaSpd:			.byte	2	;P1,Velocity feedback gain	Fixed point
GaTqP:			.byte	2	;P2,Proportional gain		Fixed point
GaTqI:			.byte	2	;P3,Integral gain		Fixed point
LimTrq:			.byte	2	;P4,Torque limit		Integer
GaEG:			.byte	2	;P5,EG feedback gain		Fixed point
Gear:			.byte	2	;P6,Gear ratio for pulsed input	8.8 fixed point
				.byte	2	;P7

UART_COMMAND: 	.byte 	1	;
UART_DATA_SIZE:	.byte 	1	;
UART_OUT_T: 	.byte 	1	;
UART_OUT_H:		.byte 	1	;
UART_IN:		.byte 	16	;
UART_OUT:		.byte	16	;

_CtDiv4:		.byte	1	;  1/5 (200 Hz) ( r25  .)
_200HzCntL:		.byte	1  	;  200 Hz 
_200HzCntH:		.byte	1  	; 
_VelocityL:		.byte	1	;    
_VelocityH:		.byte	1	;
_Mode:			.byte	1	;  -	

Temp1:			.byte	1	;
Temp2:			.byte	1	;
Temp3:			.byte	1	;

Del_Cnt1:		.byte	1	;
Del_Cnt2:		.byte	1	;

Timer_Err1:		.byte	1	;
Timer_Err2:		.byte	1	;



;----------------------------------------------------------;
; EEPROM
;----------------------------------------------------------;
.eseg
;--------- 1  -------------------
	.org	0x0010		; Memory bank 0
	.dw	500, 8960, 512, 32, 250, 1500	;  1  

	; Memory bank 1 : Tamagawa Seiki TS1410N1
	.dw	300, 0x0800, 0x0300, 0x0060, 180, 0x0550 	;

	; Memory bank 2 : Yasukawa Electric UGTMEM-A1SA51
	.dw	300, 0x0500, 0x0180, 0x0060, 255, 0x0340 	;

	; Memory bank 3 : Matsushita Electric MCN-14EAEC (6V, 40p/r)
	.dw	300, 0x09C4, 0x0640, 0x0060, 180, 0x0550	;

	; Memory bank 4 : 
	.dw	300, 0x0F00, 0x0300, 0x0060, 180, 0x0550 	;

;--------- 2  -------------------
	.org	0x0060		; Memory bank 0
	.dw	500, 8960, 512, 32, 250, 1500	;  1  

	; Memory bank 1 : Tamagawa Seiki TS1410N1
	.dw	300, 0x0800, 0x0300, 0x0060, 180, 0x0550 	;

	; Memory bank 2 : Yasukawa Electric UGTMEM-A1SA51
	.dw	300, 0x0500, 0x0180, 0x0060, 255, 0x0340 	;

	; Memory bank 3 : Matsushita Electric MCN-14EAEC (6V, 40p/r)
	.dw	300, 0x09C4, 0x0640, 0x0060, 180, 0x0550	;

	; Memory bank 4 : 
	.dw	300, 0x0F00, 0x0300, 0x0060, 180, 0x0550 	;

;--------- 3  -------------------
		.org	0x00C0		; Memory bank 0
	.dw	500, 8960, 512, 32, 250, 1500	;  1  

	; Memory bank 1 : Tamagawa Seiki TS1410N1
	.dw	300, 0x0800, 0x0300, 0x0060, 180, 0x0550 	;

	; Memory bank 2 : Yasukawa Electric UGTMEM-A1SA51
	.dw	300, 0x0500, 0x0180, 0x0060, 255, 0x0340 	;

	; Memory bank 3 : Matsushita Electric MCN-14EAEC (6V, 40p/r)
	.dw	300, 0x09C4, 0x0640, 0x0060, 180, 0x0550	;

	; Memory bank 4 : 
	.dw	300, 0x0F00, 0x0300, 0x0060, 180, 0x0550 	;


.equ	N_PARM = 6			; Number of parameters par bank.
;----------------------------------------------------------;


.cseg
		.org	$0000

	RJMP 	Start     		;   reset
							;     
							; (- )   
	RJMP 	INT0_INT		; External Interrupt Request 0
	RJMP 	Start			; External Interrupt Request 1
	RJMP 	Start			; PCINT0		; Pin Change Interrupt Request 0
	RJMP 	Start			; PCINT1		; Pin Change Interrupt Request 0
	RJMP 	Start			; PCINT2		; Pin Change Interrupt Request 1
	RJMP 	Start			; WDT			; Watchdog Time-out Interrupt
	RJMP 	Start			; TIM2_COMPA	; Timer/Counter2 Compare Match A
	RJMP 	Start			; TIM2_COMPB	; Timer/Counter2 Compare Match B
	RJMP 	Start			; RJMP 	TIMER2_OVF_INT	; Timer/Counter2 Overflow
	RJMP 	Start			; TIM1_CAPT		; Timer/Counter1 Capture Event
	RJMP 	Start			; TIM1_COMPA	; Timer/Counter1 Compare Match A
	RJMP 	Start			; TIM1_COMPB	; Timer/Counter1 Compare Match B
	RJMP 	Start			; TIM1_OVF		; Timer/Counter1 Overflow
	RJMP 	Start			; TIM0_COMPA	; TimerCounter0 Compare Match A
	RJMP 	Start			; TIM0_COMPB	; TimerCounter0 Compare Match B
	RJMP 	TIMER0_OVF_INT	; TIM_OVF	 	; Timer/Couner0 Overflow
	RJMP 	Start			; SPI_SRC		; SPI Serial Transfer Complete
	RJMP 	ReceiveInt		; USART Rx Complete
	RJMP 	TransmitInt		; USART, Data Register Empty
	RJMP 	Start			; USART_TXC		; USART Tx Complete
	RJMP	Start			; ADC Conversion Complete
	RJMP 	Start			; EE_RDY		; EEPROM Ready
	RJMP 	ANA_COMP		; ANA_COMP		; Analog Comparator
	RJMP 	Start			; TWI			; Two-wire Serial Interface
	RJMP 	Start			; SPM_RDY		; Store Program Memory Read

	
Start:
	clr		_0
	ldi		r16, low(RAMEND)
	out		spl, r16
	ldi		r16, HIGH(RAMEND)
	out		sph, r16
;-------------------------------------------------------
	wdr								;   
	lds 	r16,WDTCSR				;
	ori		r16,(1<<WDCE)|(1<<WDE)	;	
	sts 	WDTCSR,r16				;  WDT 

	ldi		r16,0b0001011			;   
	sts 	WDTCSR,r16				; t=125ms


;-----------------------------------------------------------
		;  RAM
	ldi		YL, LOW(0x100)
	ldi		YH, HIGH(0x100)

	ldi		r16, 0 					; R17=2 / 512 bytes (for AtMega48)
	ldi		r17, 4 					; R17=4 / 1024 bytes (for AtMega88 & 168);
cl_ram:
	st		Y+, _0
	sec
	sbc		r16, _0
	sbc		r17, _0
	cp		r16, _0
	cpc		r17, _0
	brne	cl_ram
	
;-----------------------------------------------------------

; --------- (  (pin)   ATMega48/88/168   TQPF-32) ------
;**********   B: **********************************************
	ldi	r16,0b11000111			; PB7(pin8) - XTAL2
	out	DDRB,r16				; PB6(pin7) - XTAL1
	ldi	r16,0b00111000			; PB5(pin17) -  Limit_R (SCK)
	out	PORTB,r16				; PB4(pin16) -  Limit_L (MISO)
 								; PB3(pin15) -  Mode_Lim (MOSI)
								; PB2(pin14) -  1 (-2)
								; PB1(pin13) -  1 (-1)
							  	; PB0(pin12) -  ( .)

;**********   : **********************************************
	ldi	r16,0b11110111			; PC7(xxx) -    
	out	DDRC,r16				; PC6(pin29) - Reset ( .)
	ldi	r16,0b00001000			; PC5(pin28) -  Direction_L 
	out	PORTC,r16				; PC4(pin27) -  Direction_R
								; PC3(pin26) -  (DIRECT (F/R) - ) 
							  	; PC2(pin25) -  Led_Error
								; PC1(pin24) -  ( .)
								; PC0(pin23) -  ( .)
	
;**********   D: **********************************************
	ldi	r16,0b00001010			; PD7(pin11) - AIN1 - (-)  
	out	DDRD,r16				; PD6(pin10) - AIN0 - (+) 
	ldi	r16,0b00110101			; PD5(pin9) -    
	out	PORTD,r16				; PD4(pin2) -    
								; PD3(pin1) -  ( .)
  								; PD2(pin32) -  Step 
					 			; PD1(pin31) -  TX 
								; PD0(pin30) -  RX 
;     "" 

;----------------------------------------------------------------
	ldi		r16,0b00011000			;  .   
	out		ACSR,r16				;       
;----------------------------------------------------------------
; timer 0 - -
	ldi		r16,0b00000011			; CK/64 prescaler	
	out		TCCR0B, r16				; (976)
;---------------------------------------------------------------
; timer1 -   
	ldi		r16,0b10100001			; Phase correct PWM, 8-bit (for 16MHz)
;	ldi		r16,0b10100010			; Phase correct PWM, 9-bit (for 20MHz)
	sts		TCCR1A, r16				;

	ldi		r16,0b00000001			;   
	sts		TCCR1B, r16
		
	sts		OCR1AH, _0
	sts		OCR1AL, _0
	sts		OCR1BH, _0
	sts		OCR1BL, _0

;--------------------------------------------------------------------
	ldi		r16,0b00000001			;   
	sts		TIMSK0,r16				;    Timer0 
;---------------------------------------------------------------------
	ldi		r16,0b00000010			;INT0    (rising ISC01 | ISC00)
	sts		EICRA, r16				;
	
	ldi		r16,0b00000001			;   INT0 (Step)
	out		EIMSK, r16				;
;-----------------------------------------------------------------

	;  UART, 38400 bps
	sts		UBRR0H, _0
	ldi		r16,25					; 25 - 16MHz, 32 - 20MHz
	sts		UBRR0L, r16

	ldi		r16,0b00011000			; |(1<<CHR9)|(1<<TXB8)
	sts		UCSR0B, r16

	ldi		r16,0b00000110 			; 8-bit/no parity/1 stop
	sts		UCSR0C, r16

   	sts 	UART_OUT_T,_0
   	sts 	UART_OUT_H,_0
;---------------------------------------------------------------------	
	cli								;   
	rcall	del_250ms				;   0,25.    

	ldi		r16,0b10011000
	sts		UCSR0B, r16
	rcall	del_250ms				;     0,25.

	;   PID-  EEPROM  
	clr		Flags2
	rcall	EEPROM_RAM_Majorization

	rcall	FirstInitServo			; ""   
	sei								;   
	sbr		_Flags, (1<<wait_cmd_bit) 	;   UART
			
		;   UART
	ldi		zl, LOW(HelloMsg*2)
	ldi		zh, HIGH(HelloMsg*2)
	rcall	WriteUARTString
	
	rcall	FirstInitServo			;	
	ldi		r16, 0
	sts		_Mode, r16				;   
	rcall	del_250ms				;   0,25
	rcall	FirstInitServo			;
	ldi		r16, 3					; 
	sts		_Mode, r16				;   



;*******************************************************************
Main:
	wdr								;   
	ldi		r16,105					;  (20)   "" 
	sts		Del_Cnt2,r16			;  ""

Del_1:
	sbrc	_Flags, command_bit
	rcall	In_Message				;
	sbrc	Flags2,Error			;    "" ()
	rjmp	Servo_Error				;  ""  ErrorMode   				

	lds		r16,Del_Cnt1			;
	dec		r16						;
	sts		Del_Cnt1,r16			;
	brne	Del_1					;
	lds		r16,Del_Cnt2			;
	dec		r16						;
	sts		Del_Cnt2,r16			;
	brne	Del_1					;

;------------------------------------

Inp_Mode_Limit:
	sbis	PINB,Mode_Lim			;      ""
	rjmp	PC+4					;     (Pin Mode_Lim=1)
	cbr		Flags2,(1<<Mode1_Lim)	;     ,
	cbr		Flags2,(1<<Mode2_Lim)	; ..     
	rjmp	Inp_Lim_L				;   - 

	sbrc	Flags2,Mode1_Lim		;    (Pin Mode_Lim=0),
	rjmp	PC+3					;      timer2  Mode1_Lim
	sbr		Flags2,(1<<Mode1_Lim)	;      ( 20)
	rjmp	Inp_Lim_L				;       ,
	sbr		Flags2,(1<<Mode2_Lim)	;      (Mode2_Lim)	
									; (  - )

Inp_Lim_L:
	sbrc	Flags2,Mode2_Lim		;        
	rjmp	PC+4					; .
	sbic	PINB,Lim_L				;      
	rjmp	lim_L01					;     Mode2_Lim, 
	rjmp	lim_L02					;    Lim_L
	sbic	PINB,Lim_L				;   ,   .
	rjmp	lim_L02					; 
lim_L01:
	cbr		Flags2,(1<<Lim1_L)		;  ""  , 
	cbr		Flags2,(1<<Lim2_L)		;     
	rjmp	Inp_Lim_R				;    
lim_L02:
	sbrc	Flags2,Lim1_L			;  "" ,    
	rjmp	PC+3					;   timer2  Lim1_L
	sbr		Flags2,(1<<Lim1_L)		;      ( 20)
	rjmp	Inp_Lim_R				;      ,
	sbr		Flags2,(1<<Lim2_L)		;     "" (Lim2_L)


Inp_Lim_R:
	sbrc	Flags2,Mode2_Lim		;        
	rjmp	PC+4					; .
	sbic	PINB,Lim_R				;      
	rjmp	lim_R01					;     Mode2_Lim, 
	rjmp	lim_R02					;    Lim_R
	sbic	PINB,Lim_R				;   ,   .
	rjmp	lim_R02					; 
lim_R01:
	cbr		Flags2,(1<<Lim1_R)		;  ""  , 
	cbr		Flags2,(1<<Lim2_R)		;     
	rjmp	EStop_Limit				;    
lim_R02:
	sbrc	Flags2,Lim1_R			;  "" ,    
	rjmp	PC+3					;   Lim1_R
	sbr		Flags2,(1<<Lim1_R)		;      ( 20)
	rjmp	EStop_Limit				;      ,
	sbr		Flags2,(1<<Lim2_R)		;     "" (Lim2_R)

EStop_Limit:
	sbrs	Flags2,Lim2_L			;     ,
	rjmp	Del_Error				; ..   EMERGENCY STOP,
	sbrs	Flags2,Lim2_R			;    / servo_error
	rjmp	Del_Error				;
	rjmp	Servo_Error				;

Del_Error:
	cbr		Flags2,(1<<Error)		;

	lds		r16, _Mode
	cpi		r16, 3					; 
	breq	d_err					;      "", 
	sts		Timer_Err1,_0			; ""     Error 
	sts		Timer_Err2,_0			;          
	rjmp	Main					;     
d_err:
	lds		r16,Timer_Err1			; 
	inc		r16						;
	sts		Timer_Err1,r16			;
	breq	PC+2					;
	rjmp	Main					;

	lds		r16,Timer_Err2			;
	inc		r16						;
	sts		Timer_Err2,r16			;
	cpi		r16,2					;     Error 
	breq	PC+2					; 
	rjmp	Main					;



Servo_Error:
	sbr		Flags2,(1<<Error)		;   Error
	sts		_Mode,_0				;   
	rcall	InitServo				; 
	sbi		PORTC,HL_Error 			; Error LED On
	clr		r19
	clr		r20						; Output = 0V
	cbi		PORTC,Dir_R				;
	cbi		PORTC,Dir_L				;
	sts		OCR1AH, _0				;
	sts		OCR1AL, _0				;
	sts		OCR1BH, _0				;
	sts		OCR1BL, _0				;
	rjmp	Del_1



;================================================================================
;  
;================================================================================
FirstInitServo:
	ldi		r16,0b00011000		;     
	out		ACSR,r16			; 
	clr		_Flags				;
	clr		Flags2				;
	clr		_PvEnc				;
	clr		_PvDir				;
	clr		zh					;
	clr		zl					;
InitServo:
	wdr							;   
	sts		CtSub+0,_0			;
	sts		CtSub+1,_0			;
	sts		CtPos+0,_0			;
	sts		CtPos+1,_0			;
	sts		CtPos+2,_0			;			
	clr		_PosL
	clr		_PosM
	clr		_PosH
	cbi		PORTC,HL_Error 		; Error LED off
 	ret



;-------------------------------------------------------------------------------
; ******************************************************************************
In_Message:
	lds		r16, UART_COMMAND

	cpi		r16, 'p' 				; Get Position
	breq	do_get_position

	cpi		r16, 'V' 				; Get Velocity and 1kHz counter
	breq	do_get_vel_1khz_cnt

	cpi		r16, 'B' 				; Set sub-command
	brne	PC+2
	rjmp	do_set_sub
			
	cpi		r16, 'P' 				; Set Position
	brne	PC+2
	rjmp	do_set_position
		
	cpi		r16, 'M' 				; Set Servo Mode
	brne	PC+2
	rjmp	do_set_servomode

	cpi		r16, 'm' 				; Get Servo Mode
	brne	PC+2
	rjmp	do_get_servomode

	cpi		r16, 'L' 				; Load from EEPROM bank
	brne	PC+2
	rjmp	do_load_eeprom_bank

	cpi		r16, 'S' 				; Save to EEPROM bank
	brne	PC+2
	rjmp	do_save_eeprom_bank
		
	cpi		r16, 'I' 				; Get Welcome string and Version
	brne	PC+2
	rjmp	do_welcome_str

	cpi		r16, 'W' 				; Write Param X
	brne	PC+2
	rjmp	do_write_param1

	cpi		r16, 'R' 				; Read param X
	brne	PC+2
	rjmp	do_read_param1

	cpi		r16, 'z' 				; Read position 2
	brne	PC+2
	rjmp	do_read_pozzz

	rjmp	cmd_fail				;  

; ++++++++++++++++++++++++++++++++++++
;  ! !
; ++++++++++++++++++++++++++++++++++++
do_get_position:
	cli
	mov		r16, _PosH
	mov		zl, _PosM
	mov		uart_r, _PosL
	sei
	rcall	MovsBuf
	mov		uart_r, zl
	rcall	MovsBuf
	mov		uart_r, r16
	rcall	MovsBuf
	rjmp	cmd_ok

; ++++++++++++++++++++++++++++++++++++
;  ! ,    1 kHz!
; ++++++++++++++++++++++++++++++++++++
do_get_vel_1khz_cnt:
	cli
	lds		uart_r, _200HzCntL
	lds		r16, _200HzCntH
	sei
	rcall	MovsBuf
	mov		uart_r, r16
	rcall	MovsBuf

	cli
	lds		r16, Velocity+1			;
	lds		uart_r, Velocity+0 		;
	sei	
	rcall	MovsBuf
	mov		uart_r, r16
	rcall	MovsBuf

	cli
	lds		r16, VelError200+1		;
	lds		uart_r, VelError200+0	;
	sei	
	rcall	MovsBuf
	mov		uart_r, r16
	rcall	MovsBuf
	rjmp	cmd_ok

; ++++++++++++++++++++++++++++++++++++
;  ! SUB-COMMAND!
; ++++++++++++++++++++++++++++++++++++
do_set_sub:
	cli
	lds		r16, UART_IN+1
	sts		CtSub+0, r16
	lds		r16, UART_IN+0
	sts		CtSub+1, r16
	sei
	rjmp	cmd_ok

; ++++++++++++++++++++++++++++++++++++
;  ! !
; ++++++++++++++++++++++++++++++++++++
do_set_position:
	cli
	lds		r16, UART_IN+2
	sts		CtPos+0, r16
	lds		r16, UART_IN+1
	sts		CtPos+1, r16
	lds		r16, UART_IN+0
	sts		CtPos+2, r16
	sei
	rjmp	cmd_ok

; ++++++++++++++++++++++++++++++++++++
;  !   EEPROM!
; ++++++++++++++++++++++++++++++++++++
do_load_eeprom_bank:
	lds		r16, UART_IN
	cpi		r16, (E2END+1)/N_PARM/2
	brcs    do_le
	rjmp	cmd_fail
do_le:	
	rcall	LoadPIDParamEEPROM
	rjmp	cmd_ok

; ++++++++++++++++++++++++++++++++++++
;  !   EEPROM!
; ++++++++++++++++++++++++++++++++++++
do_save_eeprom_bank:
	lds		r16, UART_IN
	cpi		r16, (E2END+1)/N_PARM/2
	brcs    do_se
	rjmp	cmd_fail
do_se:	
	rcall	SavePIDParamEEPROM
	rjmp	cmd_ok

; ++++++++++++++++++++++++++++++++++++
;  !  !
; ++++++++++++++++++++++++++++++++++++
do_set_servomode:
	cbr		Flags2,(1<<Error)		;
	cbr		Flags2, (1<<StepCNC)	;
	cli
	lds		r16,UART_IN+0
	sts		_Mode,r16
	rcall	InitServo
	sei
	rjmp	cmd_ok

; ++++++++++++++++++++++++++++++++++++
;  !  !
; ++++++++++++++++++++++++++++++++++++
do_get_servomode:
	lds		r16,_Mode
	mov		uart_r,r16	
	rcall	MovsBuf
	rjmp	cmd_ok

; ++++++++++++++++++++++++++++++++++++
;  !  X!
; ++++++++++++++++++++++++++++++++++++
do_read_param1:
	lds		r16, UART_IN 			;  
	cpi		r16, N_PARM  			;   ?
	brcc    cmd_fail

	ldi		r17, 2
	clr		r19
srp_a:
	subi	r16, 1
	brcs	srp_b
	add		r19, r17
	rjmp	srp_a

srp_b:  
	ldi		yl, low(Parms)
	ldi		yh, high(Parms)
	add		yl, r19
	adc		yh, _0

	ld 		uart_r, Y+
	rcall	MovsBuf
	ld 		uart_r, Y
	rcall	MovsBuf
	rjmp	cmd_ok

; ++++++++++++++++++++++++++++++++++++
;  !  X!
; ++++++++++++++++++++++++++++++++++++
do_write_param1:
	lds		r16, UART_IN+2 			;  
	cpi		r16, N_PARM 			;   ?
	brcc    cmd_fail

	ldi		r17, 2
	clr		r19
swp_a:
  	subi	r16, 1
	brcs	swp_b
	add		r19, r17
	rjmp	swp_a

swp_b:  
	ldi		yl, low(Parms)
	ldi		yh, high(Parms)
	add		yl, r19
	adc		yh, _0

	lds		r16, UART_IN+1
	st 		Y+, r16
	lds		r16, UART_IN
	st 		Y, r16
	rjmp	cmd_ok

do_welcome_str:
	ldi		zl, low(HelloMsg*2)
	ldi		zh, high(HelloMsg*2)
	rcall	WriteUARTString
	rjmp	cmd_ok

; ++++++++++++++++++++++++++++++++++++
;  !  2 (Z)!
; ++++++++++++++++++++++++++++++++++++
do_read_pozzz:
	cli
	lds		r16, CtPos+2
	lds		zl, CtPos+1
	lds		uart_r, CtPos+0
	sei
	rcall	MovsBuf
	mov		uart_r, zl
	rcall	MovsBuf
	mov		uart_r, r16
	rcall	MovsBuf
	rjmp	cmd_ok

;----------------------------------------------		
cmd_fail:
	ldi		uart_r, '-'
	rjmp	PC+2
cmd_ok:	
	ldi		uart_r, '+'
	rcall	MovsBuf

	cbr		_Flags, (1<<command_bit) ;  
	sbr		_Flags, (1<<wait_cmd_bit);   
	ret


;/////////////////////////////////////////////////////////////////////////////////////////
;     UART_OUT   uart_r
MovsBuf: 	
	push 	zl                 		;     
    push 	zh
    push 	r16
    push 	r17

;	(cbi	UCSR0B, UDRIE0)  		;   UDRE(
    lds		r17,UCSR0B     			; ,     )
	andi	r17,0b11011111
	sts		UCSR0B,r17

    lds 	zl,UART_OUT_T      		;    
    inc 	zl              		;    1
    andi 	zl,$0f         			;     15   
									;  (   16 )

    lds 	r16,UART_OUT_H  		;    
    cp  	zl,r16          		;       
    brne 	m1                 		;      (   
	                                ; ),  m1

    sbr		_Flags, (1<<full_bit)	;    ,   
	                               	;         
    rjmp 	endm               		;    


m1: sts 	UART_OUT_T,zl     		;       
	cbr		_Flags, (1<<full_bit)	; C  ,   
                                 	;      
; 	clr 	zh                 		;    - RAM
  	ldi		zh, high(UART_OUT)		;      RAM
    dec 	zl 
    andi 	zl,$0f
;	subi	zl,UART_OUT       		; 
	ldi		r17, low(UART_OUT)     	;     UART_OUT  
	add		zl,r17					;   
   	st  	z,uart_r           		;    
endm:
;	(sbi		UCSRB, UDRIE)      	;   UDRE
	lds		r17,UCSR0B
	ori		r17,0b00100000
	sts		UCSR0B,r17

    pop 	r17
    pop 	r16		               	;     
    pop 	zh
    pop 	zl
    ret


;    UDRE
TransmitInt:
	push	zl						;     
	push	zh
	push	r16
	in		r16, SREG
	push	r16
	push	r17

	lds		zl, UART_OUT_H			;    
;	clr		zh
	ldi		zh, high(UART_OUT)		;      RAM

;	(subi	zl,	-UART_OUT)			;       
	ldi		r17, low(UART_OUT)     	;     UART_OUT  
	add		zl,r17					;   

	ld		r16, z					;     
	sts		UDR0, r16				;    UART

	lds		zl, UART_OUT_H		 	;      ?	
	inc		zl
	andi	zl, $0f
	lds		zh, UART_OUT_T
	cp		zl, zh
	brne	tl						;   ,   t1			

;	(cbi		UCSR0B, UDRIE0)	 	;   ,   	
	lds		r17,UCSR0B				;    UDRE ( )
	andi	r17,0b11011111
	sts		UCSR0B,r17		
						
tl:	sts		UART_OUT_H, zl			;      
                               		;   .	
						
	pop		r17		
	pop		r16						;     
	out		SREG, r16
	pop		r16
	pop		zh
	pop		zl
	reti


;================================================================================
; UART RX INTERRUPT;
;          UART
;================================================================================
ReceiveInt:
	push	zl
	push	zh
	push	uart_r
	push	r16
	in		r16, SREG
	push	r16
	push	r17

	lds		r16, UDR0	

;	sbic	UCSR0A,DOR0
	lds		r17,UCSR0A
	sbrc	r17,DOR0					;  
	rjmp	qq	

	sbrs	_Flags, command_bit	
	rjmp	inl

qq:	
	sbr		_Flags, (1<<wait_cmd_bit)  	;   
	cbr		_Flags, (1<<wait_size_bit)

	ldi		uart_r, 'X'
	rcall	MovsBuf
	rjmp	endr

inl:	
	sbrs	_Flags, wait_cmd_bit		;  ?
	rjmp	r_no_cmd					; 

	sts		UART_COMMAND, r16			;   (2 cyc)
	cbr		_Flags, (1<<wait_cmd_bit)
	sbr		_Flags, (1<<wait_size_bit) 	;   
	rjmp	endr

r_no_cmd:
	sbrs	_Flags, wait_size_bit		;  -  
	rjmp	r_no_siz					; 

	cpi		r16, 16						;  ?
	brlo	r_s_ok						; !
	rjmp	qq 							; 

r_s_ok:	
	tst		r16
	brne	r_st		
	sbr		_Flags, (1<<command_bit)
	rjmp	endr		


r_st:
	dec		r16
	sts		UART_DATA_SIZE, r16 		;  
	cbr		_flags, (1<<wait_size_bit)
	rjmp	endr	

r_no_siz:
	lds		zl,	UART_DATA_SIZE			;    
	mov		uart_r, zl
;	clr		zh
	ldi		zh, high(UART_IN)
;	subi	zl, -UART_IN
 	ldi		r17, low(UART_IN)     		;
	add		zl,r17			
	
	st		z, r16
	tst		uart_r	
	brne	r_dec	; , 
	sbr		_Flags, (1<<command_bit)	
	rjmp	endr

r_dec:
	dec		uart_r
	sts		UART_DATA_SIZE, uart_r

endr:
	pop		r17	
	pop		r16
	out		SREG, r16
	pop		r16
	pop		uart_r
	pop		zh
	pop		zl
	reti



;================================================================================
;   PID-  EEPROM
; : r16 -  
;================================================================================
SavePIDParamEEPROM:
	ldi		r17, N_PARM*2	; -  (word)
	clr		r19
smu_a:
	subi	r16, 1			;   (r19 =   * - )
	brcs	se_co
	add		r19, r17		
	rjmp	smu_a
se_co:
	ldi		r16,0x10		;   
	sts		Temp1,r16		;
	ldi		r16,0x60		;    EEPROM
	sts		Temp2,r16		;
	ldi		r16,0xC0		;
	sts		Temp3,r16		;
	ldi		yl, low(Parms)	;
	ldi		yh, high(Parms)	;

see:
	wdr						;   
	cli						; . 
	rcall	del_EE			;
;	out		EEARH, _0		;  AtMega48 -     ; (  )
;     3     EEPROM
	sbic	EECR, EEPE
	rjmp	PC-1			;   ( )
	lds		r16,Temp1		;   1  EEPROM
	add		r19,r16			;
	out		EEARL, r19		; EEPROM address (0..255)
	sub		r19,r16			;
	ld		r16, Y			; 
	out		EEDR, r16
	sbi		EECR, EEMPE		;   EEMWE   5 
	sbi		EECR, EEPE		;    EEWE

	rcall	del_EE			;
	sbic	EECR, EEPE
	rjmp	PC-1			;   ( )
	lds		r16,Temp2		;   2  EEPROM
	add		r19,r16			;
	out		EEARL, r19		; EEPROM address (0..255)
	sub		r19,r16			;
	ld		r16, Y			; 
	out		EEDR, r16		;
	sbi		EECR, EEMPE		;   EEMWE   5 
	sbi		EECR, EEPE		;    EEWE

	rcall	del_EE			;
	sbic	EECR, EEPE
	rjmp	PC-1			;   ( )
	lds		r16,Temp3		;   3  EEPROM
	add		r19,r16			;
	out		EEARL, r19		; EEPROM address (0..255)
	sub		r19,r16			;
	inc		r19				;
	ld		r16, Y+			; 
	out		EEDR, r16		;
	sbi		EECR, EEMPE		;   EEMWE   5 
	sbi		EECR, EEPE		;    EEWE

	dec		r17
	brne	see

	ldi		r16,5			;   
	out		EEARL, r16		;  
	sei
	ret

del_EE:
	rjmp	PC+1			;
	rjmp	PC+1			;	
	rjmp	PC+1			;
	ret

;================================================================================
;   PID-  EEPROM
; : r16 -  
;================================================================================
LoadPIDParamEEPROM:
	wdr						;   
	cli						; . 
	ldi		r17, N_PARM*2 	; -  (word)
	clr		r19
lmu_a:	
	subi	r16, 1			;   (r19 =   * - )
	brcs	le_co
	add		r19, r17
	rjmp	lmu_a
le_co:	
	ldi		r16,0x10		;   
	sts		Temp1,r16		;    EEPROM	
	ldi		yl, low(Parms)
	ldi		yh, high(Parms)
wwco:
	sbic	EECR, EEPE
	rjmp	wwco
lee:	
;	out		EEARH, _0		;  AtMega48 -     ;(  )
	lds		r16,Temp1		; 
	add		r19,r16			;
	out		EEARL, r19		; EEPROM address (0..255)
	sub		r19,r16			; 
	inc		r19				;
	sbi		EECR, EERE		; Read data (EEDR)
	in		r16, EEDR		;
	st		Y+, r16			; store data to RAM
	dec		r17
	brne	lee				;

	ldi		r16,5			;   
	out		EEARL, r16		;  
	sei
	ret


;================================================================================
;   INT0 -    
;================================================================================
INT0_INT:
	cli
	push	r0
	in		r0, SREG		 	; Save carry flag
	push	zl

	sec
	lds		zl,	CtPos+0
	sbic	PINC,DIRECT 		; F/R = 0?	 
	rjmp	i0frw		     	 ; 
 
		; 
	sbc		zl, _0
	sts		CtPos+0, zl
	lds		zl, CtPos+1
	sbc		zl, _0
	sts		CtPos+1, zl
	lds		zl, CtPos+2
	sbc		zl, _0
	rjmp	i0so

i0frw:	; 
	adc		zl, _0
	sts		CtPos+0, zl
	lds		zl, CtPos+1
	adc		zl, _0
	sts		CtPos+1, zl
	lds		zl, CtPos+2
	adc		zl, _0

i0so:
	sts		CtPos+2, zl
	sts		Timer_Err1,_0	; ""     Error 
	sts		Timer_Err2,_0	;        
	pop		zl
	out		SREG, r0
	pop		r0
	sei
	reti


;================================================================================
;    ANA_COMP -  
;================================================================================
ANA_COMP:
	cli
	push	r0
	push	r16
	push	zl
	push	zh
	in		r0, SREG


	mov		zl, _PvEnc			;   zl[1:0]
		
	in		_PvEnc, PIND		;     _PvEnc[5:4]
	swap	_PvEnc				;   _PvEnc[1:0]

	ldi		zh, 1

	sbrc	_PvEnc, 1			; Skip,    A  
	eor		_PvEnc, zh			; B xor 1

		; A B  A B
		; 1 1  1 0
		; 0 0  0 0
		; 1 0  1 1
		; 0 1  0 1

	sub		zl, _PvEnc			; Decode motion (zl = zl - _PvEnc)
	andi	zl, 3				; zl = zl and b11
	breq	ana_comp_ext		; 00 -  
	cpi		zl, 1
	breq	enc_inc				; 01 -  
	cpi		zl, 3
	breq	enc_dec				; 11 -  

		; 10 - 
	mov		zl, _PvDir
	mov		zh, _PvDir			; Count by 2 for previous direction
	lsl		zl					; zl=zl shl 1, bit0 cleared
	asr		zh					; zh=zh shr 1, bit7=const, bit0 -> carry flag	
	rjmp	enc_add

enc_dec:
	ldi		zl, low(-1)
	ldi		zh, high(-1)
	rjmp	enc_setpvdir

enc_inc:
	ldi		zl, low(1)
	ldi		zh, high(1)
enc_setpvdir:
	mov		_PvDir, zl
enc_add:
	add		_PosL, zl			;  
	adc		_PosM, zh
	adc		_PosH, zh

ana_comp_ext:
	out		SREG, r0
	pop		zh
	pop		zl
	pop		r16
	pop		r0
	sei
	reti
	
	
;================================================================================
;    TIMER0 - Servo - 1kHz
;================================================================================
TIMER0_OVF_INT:
	sei
	push	r0
	push	r1
	push	r2
	push	r3
	push	r16
	push	r17
	push	r18
	push	r19
	push	r20
	push	r21
	push	r22

	ldi		zl, 6 				; b11111101 (253, $FD)
	out		TCNT0, zl			; Restore Clock divisor


	; 200 HZ
	lds		r16,_CtDiv4			;
	dec		r16					;  1/4 
	brne	no_200hz			;  0
	sts		_CtDiv4,r16			;

	;   200Hz
	sec
	lds		r16,_200HzCntL		;
	adc		r16,_0				;
	sts		_200HzCntL,r16		;	

	lds		r16,_200HzCntH		;
	adc		r16,_0				;
	sts		_200HzCntH,r16		;

	lds		r16,_VelocityL
	sts		Velocity+0,r16
	lds		r16,_VelocityH
	sts		Velocity+1,r16
		
	lds		zl, VelError+0
	sts		VelError200+0, zl
	lds		zl, VelError+1
	sts		VelError200+1, zl
		
	sts		VelError,_0
	sts		VelError+1,_0
	sts		_VelocityL,_0
	sts		_VelocityH,_0
		
	ldi		r16,5
no_200hz:
	sts		_CtDiv4,r16			;
	lds		r2, PvPos+0			;  
	lds		r3, PvPos+1
	cli		
	sub		r2, _PosL
	sbc		r3, _PosM
	sts		PvPos+0, _PosL
	sts		PvPos+1, _PosM
	sei

		;   
	lds		r16,_VelocityL
	add		r16, r2
	sts		_VelocityL,r16
	lds		r16,_VelocityH
	adc		r16, r3
	sts		_VelocityH,r16

	com		r2					; $ff - r2
	com		r3					; $ff - r3
	adc		r2, _0
	adc		r3, _0				; r3:r2 = Velocity

		;  
	lds		r16, _Mode
	cpi		r16, 3				;  ( 3)
	breq	tap_position

	lds		r0, CtSub+0
	lds		r1, CtSub+1			; S- ( 0/1/2)
	mov		zl, r0
	mov		zh, r1
	mov		r19, r0
	mov		r20, r1

	cpi		r16, 2				;  ( 2)
	breq	PC+5
	cpi		r16, 1
	brne	PC+2
	rjmp	tap_torque			;   ( 1)
	rjmp	tap_voltage			;  ( 0)
	rjmp	tap_velocity

;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
; 
;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
tap_position:
	cli
	lds		r16, CtPos+0			; 24-  
	lds		r17, CtPos+1			;
	lds		r18, CtPos+2			;
	
	sub		r16, _PosL				;   (dX)
	sbc		r17, _PosM				;
	sbc		r18, _PosH				;
	sei

;-------      ( Error) ---- 
;	sts		Temp3,r22				;
	ldi		r22,0x03				; 0x03 = 3step, (0x0A = 10step)
	cp		r16,r22					; 
	cpc		r17,_0					; 
	cpc		r18,_0					; 
	brsh	PC+2					; 
	rjmp	comp_01	

	ldi		r22,0xFC				; 0xFFFC = -3step; (0xFFF6 = -10step)
	cp		r16,r22					; 
	ldi		r22,0xFF				; 
	cpc		r17,r22					; 
	cpc		r18,r22					; 
	brlo	comp_02					;     +/-3 a, 
comp_01:
	sts		Timer_Err1,_0			; ""     Error 
	sts		Timer_Err2,_0			;          
	rjmp	comp_end				;   

comp_02:
	ldi		r22,0x21				; 0x4E21 = 20 000 step (0x2710 = 10 000 step)
	cp		r16,r22					;    20 000 ,
	ldi		r22,0x4E				;    servo_error,
	cpc		r17,r22					;     3  
	cpc		r18,_0					;  20 000, 
	brlt	comp_03					;     
	rjmp	Servo_Error				; 


comp_03:
	ldi		r22,0xE0				; 0xFFB1E0 = -20000 step (0xFFD8F0 = -10000 step)
	cp		r16,r22					;    -20 000 ,	
	ldi		r22,0xB1				;    servo_error
	cpc		r17,r22					;     -3 
	ldi		r22,0xFF				;   -20 000,
	cpc		r18,r22					;  
	brge	comp_end				;  
	rjmp	Servo_Error				; 
	rjmp	Servo_Error				; 

comp_end:
;-------------------------------------------------------------------

	lds		r19, LimSpd+0   	;  
	lds		r20, LimSpd+1
	clr		r21
	cp		r16, r19			; 
	cpc		r17, r20
	cpc		r18, r21
	brge	pVovr; PC+11

	com		r19					; $ff - r19
	com		r20					; $ff - r20
	com		r21					; $ff - r21 = $ff

	adc		r19, _0
	adc		r20, _0
	adc		r21, _0

	cp		r16, r19
	cpc		r17, r20
	cpc		r18, r21
	brge	pVovr2 				; PC+3

pVovr:
	mov		r16, r19
	mov		r17, r20
pVovr2:
	mov		r0, r16
	mov		r1, r17				; r1:r0 = velocity command


;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
; 
; r3:r2
; r1:r0 
;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
tap_velocity:
	mov		r16, r2
	mov		r17, r3			; Velocity loop gain ( )

	lds		r18, GaSpd+0
	lds		r19, GaSpd+1
	rcall	muls1616		; r20:19 - scaled velocity

	sub		r0, r19			;  
	sbc		r1, r20

		;   -  
	lds		r16, VelError+0
	lds		r17, VelError+1
	add		r16, r0
	adc		r17, r1
	sts		VelError+0, r16	;  
	sts		VelError+1, r17	;  
		

	mov		r16, r0			;  
	mov		r17, r1			;  	
	lds		r18, GaTqP+0	;
	lds		r19, GaTqP+1	;
	rcall	muls1616		; r20:19 - proportional term
	mov		zl, r19
	mov		zh, r20
		
	lds		r16, PvInt+0	;  
	lds		r17, PvInt+1	;  	
	add		r0, r16			;
	adc		r1, r17			;
	mov		r16, r0			;
	mov		r17, r1			;
	lds		r18, GaTqI+0	;
	lds		r19, GaTqI+1	;
	rcall	muls1616		; r20:r19 - integral term
	add		zl, r19			;
	adc		zh, r20			; z:=proportional term + integral term

	lds		r18, LimTrq+0	;  
	ldi		r19, 0			;
	cp		zl, r18			;
	cpc		zh, r19			;
	brge	PC+6			;
	neg		r18				; $00 - r18
	com		r19				; $ff - r19
	cp		zl, r18			;
	cpc		zh, r19			;
	brge	PC+4			;
	mov		zl, r18			;
	mov		zh, r19			;
	rjmp	PC+5			;
	sts		PvInt+0, r0		; Anti-wind-up
	sts		PvInt+1, r1		;

;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
; 
;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
tap_torque:
	mov		r16, r2
	mov		r17, r3			; EG compensation
	lds		r18, GaEG+0		;
	lds		r19, GaEG+1		;
	rcall	muls1616		;
	add		r19, zl			;
	adc		r20, zh			; r20:r19 - Voltage command

;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
; 
; r20:r19
;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
tap_voltage:
	ldi		r16, low(240)		;
	ldi		r17, high(240)		;   -240  +240.
	cp		r19, r16			;
	cpc		r20, r17			;
	brge	t6 					;		

	ldi		r16, low(-240)		;
	ldi		r17, high(-240)		;
	cp		r19, r16			;
	cpc		r20, r17			;
	brge	mo 					;

t6:	mov		r19, r16			;
	mov		r20, r17			;/ r20:r19

mo:
	tst		r19					;  ,     stop
	breq	StopMove			;
	tst		r20					; Motor output. Set PWM register and polarity
	brmi	Direct_R			;
	
; -----    for 16MHz, PWM =31KHz
Direct_L:			; ********** Direction_L ***************
	sbrc	Flags2,Lim2_L		;  ()   
	rjmp	StopMove			;
	sts		OCR1BL, _0			;  "" ,  	
	sts		OCR1AL, r19			; 
	cbi		PORTC,Dir_R			; 
	sbi		PORTC,Dir_L			;
	rjmp	ep 			

Direct_R:			; ********** Direction_R **************
	sbrc	Flags2,Lim2_R		;  ()   
	rjmp	StopMove			;
	neg		r19					;
	sts		OCR1AL, _0			;  "" ,  
	sts		OCR1BL, r19			; 
	cbi		PORTC,Dir_L			; 
	sbi		PORTC,Dir_R			;
	rjmp	ep					;

/*
; -----    for 20MHz, PWM = 19,5KHz
Direct_L:			; ********** Direction_L *************** 
	sbrc	Flags2,Lim2_L		;  ()   
	rjmp	StopMove			;

	lsl		r19					;      
	adc		r16,_0				;  9-,    2

	sts		OCR1BH, _0			;
	sts		OCR1BL, _0			;  "" ,  	
	sts		OCR1AH, r16	
	sts		OCR1AL, r19			; 
	cbi		PORTC,Dir_R			; 
	sbi		PORTC,Dir_L			;
	rjmp	ep 			

Direct_R:			; ********** Direction_R **************
	sbrc	Flags2,Lim2_R		;  ()   
	rjmp	StopMove			;
	neg		r19					;

	lsl		r19
	adc		r16,_0

	sts		OCR1AH, _0			;
	sts		OCR1AL, _0			;  "" ,  	
	sts		OCR1BH, r16	
	sts		OCR1BL, r19			; 

	cbi		PORTC,Dir_L			; 
	sbi		PORTC,Dir_R			;
	rjmp	ep					;
; -------
*/

StopMove:
	sts		Timer_Err1,_0		; ""     Error 
	sts		Timer_Err2,_0		;           
	cbi		PORTC,Dir_R			;
	cbi		PORTC,Dir_L			;
	sts		OCR1AH, _0
	sts		OCR1AL, _0
	sts		OCR1BH, _0
	sts		OCR1BL, _0

ep:	pop		r22
	pop		r21
	pop		r20
	pop		r19
	pop		r18
	pop		r17
	pop		r16
	pop		r3
	pop		r2
	pop		r1
	pop		r0
	reti


; == END SERVO 1kHz ================


;================================================================================
;  16*16 (signed)
; : 32- (r21:r20:r19:r18)
; : 16- (r19:r18)
; : 16- (r17:r16)
;================================================================================
muls1616:
	clt
	tst		r17
	brpl	PC+6
	set
	com		r16
	com		r17
	adc		r16,_0
	adc		r17,_0

	sub		r20,r20	; clear high 16bit.
	sub		r21,r21	;
	ldi		r22,17	; r22 = loop count
	brcc	PC+3	; ---- calcurating loop
	add		r20,r16	;
	adc		r21,r17	;
	ror		r21		;
	ror		r20		;
	ror		r19		;
	ror		r18		;
	dec		r22		; if (--r22 > 0)
	brne	PC-8	;  continue loop;

	brtc	mulret ;PC+9	; Negate the result if multiplyer was negative.
	com		r18
	com		r19
	com		r20
	com		r21
	adc		r18, _0
	adc		r19, _0
	adc		r20, _0
	adc		r21, _0
mulret:	ret



;**************************************************************************************
;======================================================================================
;**********  PID -  EEPROM  RAM c  *************
;======================================================================================
EEPROM_RAM_Majorization:
	cli							;    
	wdr							;   
	ldi		r17,0x10			;   
	ldi		r18,0x60			;    EEPROM
	ldi		r19,0xC0			;
	dec		r17					;
	dec		r18					;
	dec		r19					;
	ldi		yl, low(Parms)		;
	ldi		yh, high(Parms) 	; 
	ldi		r20, N_PARM*10		; -  (word)

EE_Read:
 	inc		r17					;   
	mov		r16,r17				;   	
	rcall	EEPROM_RD			;   EEPROM
	sts		Temp1,r16			;    W   

 	inc		r18					;
	mov		r16,r18				;
	rcall	EEPROM_RD			; 
	sts		Temp2,r16			; 

  	inc		r19					;
	mov		r16,r19				;
	rcall	EEPROM_RD			; 
	sts		Temp3,r16			; 


	rcall	Srednee				;       
	lds		r16,Temp1			;  EEPROM      
	st		Y+, r16				; 


	sbrs	Flags2,EE_WR1		;      EEPROM
	rjmp	PC+4				;    ,    4 
	mov		r16,r17				;  ,      EEPROM
	rcall	EEPROM_WR			;     Temp1  EEPROM
	cbr		Flags2,(1<<EE_WR1)	;

	sbrs	Flags2,EE_WR2		;     
	rjmp	PC+4				; emp1  
	mov		r16,r18				;   EEPROM
	rcall	EEPROM_WR			;
	cbr		Flags2,(1<<EE_WR2)	; 
	
	sbrs	Flags2,EE_WR3		; 
	rjmp	PC+4				; 
	mov		r16,r19				; 
	rcall	EEPROM_WR			;
	cbr		Flags2,(1<<EE_WR3)	; 

	dec		r20					;      EEPROM   
	brne	EE_Read				;       (   
	ldi		r16,5				; EEPROM   ,    )
	rcall	EEPROM_RD			;  ""  EEPROM (  
	clr		Flags2				;  EEPROM)
	sts		Temp1,_0			;      EEPROM
	sts		Temp2,_0			;
	sts		Temp3,_0			;     / 
			
	ret							;   /

;-------------------------------------------------------------------------------------
;********** /      3  *******************
Srednee:
	clr		Flags2				;      EEPROM
	
	lds		W,Temp1				;  Temp1  2 
	lds		r16,Temp2			;
	cp		W,r16				;
	breq	srd_03				;   ,     srd_03		


	lds		r16,Temp3			;  Temp1  2    , 
	cp		W,r16				;  Temp1  3
	brne	srd_01				;    ,     srd_01
	sbr		Flags2,(1<<EE_WR2)	;  Temp1  3 ,    
	ret							;   2 . EEPROM    /

srd_01:
	lds		W,Temp2				;  Temp1  3    
	cp		W,r16				;  Temp2  3 ,   ,
	breq	srd_02				;     srd_02
	sts		Temp1,W				;         ,
	sbr		Flags2,(1<<EE_WR1)	;    Temp1  2 (  
	sbr		Flags2,(1<<EE_WR3)	; "" ),   
	ret							;   1  3     /

srd_02:
	sts		Temp1,r16			;  2  3 ,    1, 	
	sbr		Flags2,(1<<EE_WR1)	;  3  1,  
	ret							;    1 .EEPROM    /

srd_03:
	lds		W,Temp3				;  Temp1  2 ,   Temp2  3 
	cp		W,r16				;      ,  
	brne	srd_04				;   /    
	ret							;   EEPROM
srd_04:
	sbr		Flags2,(1<<EE_WR3)	;  2  3  ,    	
	ret							;    3 . EEPROM    /



;================================================================================
;   PID-  EEPROM
; : r16 -  
; : r16 -   EEPROM
;================================================================================
EEPROM_RD:
	wdr							;   
	cli							;   
	sbic	EECR, EEPE			;     
	rjmp	PC-1				; ( )
	out		EEARH, _0			;  AtMega48 -     ; (  )
	out		EEARL, r16			; EEPROM address (0..255)
	sbi		EECR, EERE			; Read data (EEDR)
	in		r16, EEDR			;    EEPROM -> r16
	ret							;    /


;================================================================================
;   PID-  EEPROM
; : r16 -  
;================================================================================
EEPROM_WR:
	wdr							;   
	cli							;   
	sbic	EECR, EEPE	
	rjmp	PC-1				;   ( )
	out		EEARH, _0			;  AtMega48 -     ; (  )
	out		EEARL, r16			; EEPROM address (0..255)
	lds		r16, Temp1			;    Temp1
	out		EEDR, r16			;  EEDR 
	sbi		EECR, EEMPE			;   EEMPE   5 
	sbi		EECR, EEPE			;    EEPE
	ret							;    /
;================================================================================



;================================================================================
;     
;================================================================================
del_2s:
	ldi		r16,120				;
	rjmp	del_m				;

del_1s:
	ldi		r16,60				;
	rjmp	del_m				;

del_500ms:
	ldi		r16,30				;
	rjmp	del_m				;

del_250ms:
	ldi		r16,15				; 
	rjmp	del_m				;

del_100ms:
	ldi		r16,6				;
del_m:
	wdr							;   
	dec 	r17
	brne	del_m
	dec 	r18
	brne 	del_m
	dec 	r16
	brne	del_m
	ret


;================================================================================
;    UART (Input - Z)
;================================================================================
WriteUARTString:
	lpm
	adiw	zl, 1
	tst		r0
	breq	wur
	mov		uart_r, r0
	rcall	MovsBuf
	rjmp	WriteUARTString
wur:
	ret



HelloMsg:
	.db		"SERVO V3.6 M168", 0




